/*
 * critsec.c - handling of critical sections, especially for 
 * the OpenSSL cryptographic library.
 * MRR  2012-09-04
 * Licensed under the terms of the "Simplified BSD License".
 */
#include "config.h"

#ifdef HAVE_LIBCRYPTO
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/crypto.h>
#else
#error HAVE_LIBCRYPTO is not defined - probable CMake error
#endif

#include "libssh/critsec.h"

#if defined(_WIN32)
#include <windows.h>
#else
#include <pthread.h>
#endif

/* Array of synchronization objects used in code below.
 * Note: we assume that CRYPTO_num_locks() returns CRYPTO_NUM_LOCKS. 
 */
#if defined(_WIN32)
CRITICAL_SECTION critCrypt[CRYPTO_NUM_LOCKS];
#else
pthread_mutex_t mutexCrypt[CRYPTO_NUM_LOCKS];
#endif

/**
 * @brief This is the locking callback used by the OpenSSL cryptographic library,
 * as described at http://www.openssl.org/docs/crypto/threads.html.
 */
void libssh_libcrypto_locking_callback(int mode,int type, const char* file, int line)
{
	if (mode & CRYPTO_LOCK)	{
#if defined(_WIN32)
		EnterCriticalSection(&critCrypt[type]);
#else
		pthread_mutex_lock(&mutexCrypt[type]);
#endif
	} else {
#if defined(_WIN32)
		LeaveCriticalSection(&critCrypt[type]);
#else
		pthread_mutex_unlock(&mutexCrypt[type]);
#endif
	}
}

/**
 * Returns an ID that can be used the OpenSSL cryptographic library
 * to identify the current thread.
 */
unsigned long libssh_libcrypto_id_function(void)
{
	unsigned long threadID;
#ifdef _WIN32
	threadID = (unsigned long) GetCurrentThreadId();
#else
	threadID = (unsigned long) pthread_self();
#endif
	return threadID;
}

/**
 *  Initialize the critical sections used by the third-party crypto library.
 */
void libssh_init_critsec_crypto()
{
#ifdef HAVE_LIBCRYPTO
	int i;
	for (i=0;i<CRYPTO_num_locks();++i) {
#if defined(_WIN32)
		InitializeCriticalSection(&critCrypt[i]);
#else
		pthread_mutexattr_t mta;
		pthread_mutexattr_init(&mta);
		pthread_mutex_init(&mutexCrypt[i], &mta);
#endif
	}
	CRYPTO_set_id_callback(libssh_libcrypto_id_function);
	CRYPTO_set_locking_callback(libssh_libcrypto_locking_callback);
#endif
}
